# student faces submodule

# TODO: FaceFeatureStage should use RangeDictionary

import os
import vector
from random import sample, randint
# TODO: import all constants instead of this specific one?
from pygame import image, transform, BLEND_MULT


log = False

face_art_path = 'art/face/'

eye_path = face_art_path + 'eyes/'
pupil_path = face_art_path + 'pupils/'
brow_path = face_art_path + 'brow/'
mouth_path = face_art_path + 'mouth/'
hair_path = face_art_path + 'hair/'
head_path = face_art_path + 'head/'

art_scale = 0.5

BLANK = (0, 0, 0, 0)


def trim_image(image):
    "blanks (alpha 0) pixels at edges of specified image"
    # ugly, but it works
    # (possible alternate solution: export process border is
    # pure magenta (255, 0, 255), we could blank every pixel
    # of that color in the image)
    w = image.get_width()
    h = image.get_height()
    for x in range(w):
        image.set_at((x, 0), BLANK)
        image.set_at((x, 1), BLANK)
        image.set_at((x, h), BLANK)
        image.set_at((x, h - 1), BLANK)
        image.set_at((x, h - 2), BLANK)
    for y in range(h):
        image.set_at((0, y), BLANK)
        # scaling causes extra pixels of badness
        image.set_at((1, y), BLANK)
        image.set_at((w, y), BLANK)
        image.set_at((w - 1, y), BLANK)
        image.set_at((w - 2, y), BLANK)
    return image


class FaceFeatureStage():
    "specific frame of a FaceFeature, shown for a certain affinity range"
    def __init__(self, image_file, (aff_min, aff_max)):
        if log: print "loading FaceFeatureStage from %s" % image_file
        self.image = image.load(image_file)
        if not self.image:
            print "error: %s is none!" % image_file
            return
        # trim 1-pixel magenta border from export process
        self.image = trim_image(self.image)
        # scale art down from its native size
        w, h = self.image.get_size()
        w *= art_scale
        h *= art_scale
        self.image = transform.scale(self.image, (int(w), int(h)))
        # sort min/max properly in case there's a data error
        if aff_min > aff_max:
            self.min = aff_max
            self.max = aff_min
        else:
            self.min = aff_min
            self.max = aff_max
    
    def __contains__(self, aff):
        return aff >= self.min and aff <= self.max


class FaceFeature():
    "facial feature of a given style, eg blue eyes, red hair"
    # max_offset: how far this feature can drift from its origin
    # to follow the student's view direction
    max_offset = 0
    # stage mappings: keys are a filename (minus png) and a
    # tuple of the min and max affinity in which a stage is
    # appropriate.
    # worth noting that the string keys here are the filenames
    # (minus extension) that will be loaded
    stage_mapping = {
        # (this mapping is default and not meant for actual use)
        'DEBUG' : (-1000, 1000)
    }
    
    def __init__(self, stages_dir):
        stages_dir = os.path.abspath(stages_dir)
        if log: print "loading a new %s from %s" % (self.__class__, stages_dir)
        self.stages = []
        # crawl the specific dir for images that fit
        # what's in our stage mapping dict
        for f in os.listdir(stages_dir):
            if os.path.isdir(f):
                continue
            if not f.lower().endswith('.png'):
                continue
            base = os.path.basename(f)
            base = os.path.splitext(f)[0]
            if base in self.stage_mapping:
                # add to this feature's stages
                (aff_min, aff_max) = self.stage_mapping[base]
                f = stages_dir + '/' + f
                stage = FaceFeatureStage(f, (aff_min, aff_max))
                self.stages.append(stage)
        self.offset_x = 0
        self.offset_y = 0
        if log: print "FaceFeature created with stages: %s" % self.stages
    
    def dump(self):
        "debug function to show all the data about our stages"
        print "%s with stages:" % self
        for s in self.stages:
            print "image=%s, min=%s, max=%s" % (s.image, s.min, s.max)
    
    def get_aff_image(self, aff):
        for s in self.stages:
            if aff in s:
                return s.image
    
    def draw(self, screen, loc, target_dir, aff):
        if self.max_offset != 0:
            v = vector.vectorMA(loc, target_dir, self.max_offset)
            loc = v
        img = self.get_aff_image(aff)
        screen.blit(img, loc.int_tuple())


# common mapping used by many features
gen_map = {
    'hap01': (0.1, 2),
    'hap02': (2, 4),
    'hap03': (4, 6),
    'hap04': (6, 1000),
    'meh00': (0, 0),
    'mad01': (-2, -0.1),
    'mad02': (-5, -2),
    'mad03': (-8, -5),
    'mad04': (-1000, -8)
}


class Eyes(FaceFeature):
    max_offset = 8
    stage_mapping = gen_map
    
    # because pupils need to be blitted onto the eyes to look right,
    # they are not drawn directly but referred to here by Eyes.draw()
    
    def draw(self, screen, loc, target_dir, aff, pupils):
        v = vector.vectorMA(loc, target_dir, self.max_offset)
        pupil_v = vector.vectorMA(loc, target_dir, pupils.max_offset)
        pupil_v = pupil_v - v
        img = self.get_aff_image(aff).copy()
        pupil_img = pupils.get_aff_image(aff)
        if not pupil_img:
            print "couldn't get a pupil image for affinity %s from:" % aff
            pupils.dump()
        img.blit(pupil_img, pupil_v.int_tuple(), None, BLEND_MULT)
        screen.blit(img, v.int_tuple())


class Pupils(FaceFeature):
    max_offset = 10
    stage_mapping = gen_map
    # one fewer image for pupils
    stage_mapping.pop('hap04')
    stage_mapping['hap03'] = (4, 1000)


class EyeBrows(FaceFeature):
    max_offset = 8
    stage_mapping = gen_map


class Mouth(FaceFeature):
    max_offset = 7
    stage_mapping = gen_map


aff_all = (-1000, 1000)


class Head(FaceFeature):
    stage_mapping = { 'head' : aff_all }


class HairBack(FaceFeature):
    stage_mapping = { 'back' : aff_all }


class HairTop(FaceFeature):
    max_offset = 2
    stage_mapping = { 'top' : aff_all }


class HairFront(FaceFeature):
    max_offset = 4
    stage_mapping = { 'front' : aff_all }


class FaceStyle():
    "visual appearance definition for a student's face"
    
    def __init__(self, eyes, pupils, brows, mouth, head, hair_back, hair_top,
                 hair_front):
        self.eyes = eyes
        self.pupils = pupils
        self.brows = brows
        self.mouth = mouth
        self.head = head
        self.hair_back = hair_back
        self.hair_top = hair_top
        self.hair_front = hair_front
        self.thumbnail = self.init_thumbnail()
    
    def init_thumbnail(self):
        "renders a neutral image that can be used as a thumbnail"
        # start with lowest layer, use same order as draw())
        s = self.hair_back.get_aff_image(0).copy()
        s.blit(self.head.get_aff_image(0), (0, 0))
        s.blit(self.hair_top.get_aff_image(0), (0, 0))
        # get pupils clipped to eyes as in normal draw()
        eye_img = self.eyes.get_aff_image(0).copy()
        pupil_img = self.pupils.get_aff_image(0).copy()
        eye_img.blit(pupil_img, (0, 0), None, BLEND_MULT)
        s.blit(eye_img, (0, 0))
        s.blit(self.mouth.get_aff_image(0), (0, 0))
        s.blit(self.hair_front.get_aff_image(0), (0, 0))
        s.blit(self.brows.get_aff_image(0), (0, 0))
        return s
    
    def scaled_thumbnail(self, scale_x, scale_y):
        # returns our thumbnail, scaled to the specified amount
        (w, h) = self.thumbnail.get_size()
        w *= scale_x
        h *= scale_y
        return transform.smoothscale(self.thumbnail, (int(w), int(h)))
    
    def draw(self, screen, loc, target_dir, aff):
        # draw order:
        # back of hair -> head -> top of hair ->
        # eyes (with pupils) + mouth -> front of hair -> eyebrows
        self.hair_back.draw(screen, loc, target_dir, 0)
        self.head.draw(screen, loc, target_dir, 0)
        self.hair_top.draw(screen, loc, target_dir, 0)
        # only mouth and eyes reflect affinity
        # special case: eyes need their pupils passed to them for drawing
        self.eyes.draw(screen, loc, target_dir, aff, self.pupils)
        # (pupils are not drawn directly)
        self.mouth.draw(screen, loc, target_dir, aff)
        self.hair_front.draw(screen, loc, target_dir, 0)
        self.brows.draw(screen, loc, target_dir, aff)


def load_features_from_dir(dir, feature_class=FaceFeature):
    "returns a list of FaceFeatures loaded from the specified dir"
    feature_list = []
    dir = os.path.abspath(dir)
    if log: print "loading all %s in %s" % (feature_class, dir)
    for d in os.listdir(dir):
        if os.path.isfile(d):
            if log: print "%s is a file, skipping" % d
            continue
        # skip svn dirs to save time
        if d == '.svn':
            continue
        d = dir + '/' + d
        if log: print "found a feature set in %s" % d
        feature = feature_class(d)
        # if there were no valid stages for the feature
        # in this dir, don't add it to the list
        if len(feature.stages) > 0:
            feature_list.append(feature)
    return feature_list


class FaceFeatureDirectory():
    "a master list of all FaceFeatures"
    
    def __init__(self):
        # crawl all subdirs to build lists of features -
        # feature stages are handled in the constructor
        # of each feature
        # pass feature subclass to load process so the correct
        # mappings etc will be used
        print "building FaceFeatureDirectory..."
        self.pupils = load_features_from_dir(pupil_path, Pupils)
        self.eyes = load_features_from_dir(eye_path, Eyes)
        self.brows = load_features_from_dir(brow_path, EyeBrows)
        self.mouths = load_features_from_dir(mouth_path, Mouth)
        self.hair_backs = load_features_from_dir(hair_path, HairBack)
        self.hair_tops = load_features_from_dir(hair_path, HairTop)
        self.hair_fronts = load_features_from_dir(hair_path, HairFront)
        self.heads = load_features_from_dir(head_path, Head)
    
    def random_style(self):
        "returns a random style pulled from the elements of this directory"
        eyes = sample(self.eyes, 1)[0]
        pupils = sample(self.pupils, 1)[0]
        brows = sample(self.brows, 1)[0]
        mouth = sample(self.mouths, 1)[0]
        head = sample(self.heads, 1)[0]
        # same random # for matching hair pieces
        i = randint(0, len(self.hair_backs) - 1)
        hair_back = self.hair_backs[i]
        hair_top = self.hair_tops[i]
        hair_front = self.hair_fronts[i]
        return FaceStyle(eyes, pupils, brows, mouth, head, hair_back, hair_top,
                         hair_front)
