# POPULAR note passing prototype 1

# do reaction pips during note pass
# after note pass, player can mouseover/click to see reaction in depth



import os.path
import math
import pygame

# project specific
import ui
import thoughts

from colors import *
from random import randint, sample
from sys import exit

FULLSCREEN = False
SCREEN_SIZE = SCREEN_WIDTH, SCREEN_HEIGHT = 1024, 768
FRAMERATE = 30

#
# visual config
#

FONT_SIZE = 20
SMALL_FONT_SIZE = 14

#
# game vars
#

# this determines how many students there are
DESK_ROWS = 4
DESK_COLUMNS = 4
# spot player sits at in class
PLAYER_DESK = 8

# self-esteem = student's affinity for self
DEFAULT_SELF_ESTEEM = 2
# at start of game, affinity will never be more extreme (+/-) than this
# provides very basic randomization of terrain before clique formation
START_AFFINITY_MAX = 2
# run this many AI turns before game starts to swizzle network
START_AFFINITY_ITERATIONS = 0
# when we define friend groups on startup, these define peaks and variance
INITIAL_CLIQUES = 3
# exclusivity: on average, every member must like someone this much for
# membership (pay attention to how close to CLIQUE_AFFINITY_BUMP this gets
# or members will be auto-ejected from cliques)
CLIQUE_EXCLUSIVITY = 3
CLIQUE_MAX_SIZE = 4
# members start off liking each other this much
CLIQUE_AFFINITY_BUMP = 5
CLIQUE_AFFINITY_VARIANCE = 2

# personality variables
# every student will be some +/- amount of these
#
# forgiveness - do we like enemies if they begin to like our friends?
# -forgiving = vindictive
FORGIVENESS_RANGE = 3
# fickleness - how dramatically do we change our opinions of people?
# -fickle = stubborn
FICKLENESS_RANGE = 1
# gullibility - how easily we believe untruthful ideas
# -gullible = skeptical
GULLIBILITY_RANGE = 1

log = True

testing_reaction_dialog = False

# read in config file, which potentially sets any of the values above
CFG_FILENAME = 'popular.cfg'
if os.path.exists(CFG_FILENAME):
    exec(file(CFG_FILENAME))

student_names = [
    'Alice',
    'Bob',
    'Cindy',
    'Dave',
    'Eliza',
    'Fred',
    'George',
    'Helen',
    'Ingrid',
    'Jim',
    'Karla',
    'Leo',
    'Mike',
    'Nancy',
    'Otto',
    'Pauline',
    'Quincy',
    'Rick',
    'Susie',
    'Tom',
    'Uma',
    'Vernon',
    'Wilma',
    'Xavier',
    'Yolanda',
    'Zed'
]

# populate rest of game with bots in case we need them
for i in range(len(student_names) + 1, 100):
    student_names.append('ROBOT XJ-%s' % i)


#
# sim classes
#

def random_gossip(spk=None):
    "returns a randomly generated bit of gossip"
    if not spk:
        speaker = sample(students, 1)[0]
    else:
        speaker = spk
    subject = sample(students, 1)[0]
    object = sample(students, 1)[0]
    likes = bool(randint(0, 1))
    return Gossip(speaker, subject, likes, object)


class Gossip():
    "an item of gossip to be fed into the social network"
    def __init__(self, speaker, subject, likes, object):
        self.speaker = speaker
        self.subject = subject
        self.likes = likes
        self.object = object
        if log:
            print '--------------------'
            print str(self)
            print '--------------------'
    
    def __contains__(self, student):
        return student in [self.speaker, self.subject, self.object]
    
    def __str__(self):
        if self.likes:
            verb = 'likes'
        else:
            verb = 'hates'
        s = "GOSSIP: %s says: %s %s %s!" % (self.speaker.name,
            self.subject.name, verb, self.object.name)
        return s


class Reaction():
    
    "describes a student's reaction to a piece of gossip"
    
    def __init__(self, s, g):
        self.student = s
        self.gossip = g
        # dict with affinity change for each participant in the gossip
        self.aff_changes = { g.speaker : 0,
              g.subject : 0,
              g.object : 0 }
        self.truthiness = 0
        # store personality traits that manifest in this reaction
        self.forgiving = False
        self.vindictive = False
        self.gullible = False
        self.skeptical = False
        #
        # epic reaction determination algorithm
        #
        if log:
            print "%s hears %s..." % (s, g)
            s.log_affinity(g)
        # disregard your own gossip
        if g.speaker is s:
            if log:
                print "%s ignores their own gossip." % s
            return
        #
        # calculate "gut reaction" first, then "consider source"
        #
        # difference between students will be divided based on fickleness
        # higher number == smaller change
        divisor = 4 + s.fickleness
        subj_aff = s.affinity[g.subject]
        obj_aff = s.affinity[g.object]
        if g.likes:
            direction = 1
        else:
            direction = -1
        # gut reaction: person I feel more strongly about "pushes"
        # opinion of person I feel less strongly about; if equal
        # split the difference
        pusher = None
        both_push = False
        if abs(subj_aff) > abs(obj_aff):
            pusher = g.subject
            change = (subj_aff - obj_aff) / divisor * direction
            self.aff_changes[g.object] = change
        elif abs(subj_aff) < abs(obj_aff):
            pusher = g.object
            change = (obj_aff - subj_aff) / divisor * direction
            self.aff_changes[g.subject] = change
        else:
            change = (obj_aff - subj_aff) / (divisor * 2) * direction
            self.aff_changes[g.subject] = change
            self.aff_changes[g.object] = change
        # get human-readable gut reaction using thoughts.gut_reactions
        if log: print "%s's gut reaction: \"%s\"" % (s, self.thoughts())
        if log: self.log()
        
        # consider the source and the gossip's truthiness
        # (we like people for telling us truthy things, and we're
        # more inclined to "believe" things friends say)
        spk_aff = s.affinity[g.speaker]
        truth = self.get_truthiness(s, g)
        # remember truth from before we add speaker affinity
        first_truth = truth
        if log:
            print 'truthiness: %s' % truth
        
        pair = (s, g.speaker, spk_aff)
        if truth > 0:
            if log: print '%s: %s (%s) speaks the truth.' % pair
            #if g.speaker in s.enemies()
        elif truth < 0:
            if log: print '%s: %s (%s) is spreading lies!' % pair
        
        # balance truth with speaker affinity - a friend can convince us of
        # something untruthy; "doubt" truth coming from an enemy
        truth += spk_aff
        if log: print 'truth adjusted for speaker: %s' % truth
        
        # hearing lies from an enemy makes us hate them more
        if first_truth < 0 and spk_aff < 0:
            truth = first_truth
        
        # how about when friends lie to us?
        # TODO: consider having this apply to untruths from enemies as well
        if first_truth < 0 and spk_aff > 0:
            # check gullibility / skepticism
            # TODO: if we're skeptical, do we like friend less for this?
            if s.gullibility > 0:
                self.gullible = True
            elif s.gullibility < 0:
                self.skeptical = True
        
        # to normalize truth, divide by the person we feel strongest about
        # (out of all people we know)
        bfs = s.best_friends()
        bff = 0
        if len(bfs) > 0:
            bff = s.affinity[bfs[0]]
        wes = s.worst_enemies()
        worst = 0
        if len(wes) > 0:
            worst = s.affinity[wes[0]]
        self_esteem = s.affinity[s]
        passion = max(bff, abs(worst), abs(self_esteem))
        if truth:
            truth = float(truth) / passion
        if log: print 'truth adjusted for passion: %s' % truth
        # store truthiness in this class for ui etc
        self.truthiness = truth
        
        # use truthiness + speaker affinity as a bonus
        # on two things: total impact of subject/object affinity change,
        # and affinity change for speaker
        
        # apply final truth value to speaker affinity and gut reaction result
        
        # TODO: this should NOT affect person we feel more strongly about?
        # eg a truthy statement about -4 hating -5 should not increase our
        # affinity for the -5!
        self.aff_changes[g.speaker] = truth
        # FIXME: this should only apply in some cases - determine them
        if both_push:
            self.aff_changes[g.subject] += truth
            self.aff_changes[g.object] += truth
        elif pusher is g.subject:
            self.aff_changes[g.object] += truth
        elif pusher is g.object:
            self.aff_changes[g.subject] += truth
        
        # special case: if we're the subject of the gossip, we "know"
        # it is / isn't true, change only our opinion of the speaker
        if s is g.subject:
            if log: print '%s: this gossip is about me!' % s
            self.aff_changes[g.subject] = 0
            self.aff_changes[g.object] = 0
            # however truthy gossip is, we like speaker that much
            self.aff_changes[g.speaker] += truth
            if truth > 0:
                pass
                # TODO:
                # reinforce existing affinities if truthy?
                # if g.likes and g.object in s.friends():
                #     self.aff_changes[g.object] += 1
                #     if log: print '%s: That reminds me, I do like %s.' % (s, g.object)
                # elif not g.likes and g.object in s.enemies():
                #     self.aff_changes[g.object] -= 1
        
        # round to keep results meaningful and readable
        for change in self.aff_changes:
            self.aff_changes[change] = int(round(self.aff_changes[change]))
            # FIXME: for now, nullify all changes to listener self-esteem
            # (wacky side effects)
            if change is self.student:
                self.aff_changes[change] = 0
        
        if log: self.log()
        if log: print '----------------------------------------------------'
    
    def get_truthiness(self, s, g):
        """
        returns a positive or negative number indicating the 'truthiness'
        of the specified gossip.
        truthiness = emotional truth, ie how much does the gossip conform
        to our desires.
        in general, we want our friends to be liked, and our enemies to
        either reform or be disliked.
        """
        # keep student and gossip var names short, code is a mess
        # when we use self.everything
        sub_aff = s.affinity[g.subject]
        obj_aff = s.affinity[g.object]
        same_side = (sub_aff >= 0 and obj_aff >= 0) or \
            (sub_aff <= 0 and obj_aff <= 0)
        
        # usually, just consider the distance between subject and object
        d = sub_aff - obj_aff
        # if two friends or two enemies, diff = combined magnitudes
        if (g.subject in s.friends() and g.object in s.friends()) or \
                (g.subject in s.enemies() and g.object in s.enemies()):
            d = sub_aff + obj_aff
        
        t = abs(d)
        # in general, truthful = friends liked, enemies disliked
        if g.object in [self] + s.friends():
            # enemy likes friend: nooo
            if g.subject in s.enemies() and g.likes:
                t *= -1
                # check forgiveness
                # forgiving people want even enemies to like friends
                if s.forgiveness > 0:
                    if log: print '%s is forgiving.' % s
                    # TODO: enable and balance actual change
                    #t += change
                    # remember that personality modified this
                    self.forgiving = True
                elif s.forgiveness < 0:
                    # vindictive people hate enemies for liking friends
                    if log: print '%s is vindictive.' % s
                    #t -= change
                    self.vindictive = True
            if not g.likes:
                t *= -1
        elif g.object in s.enemies():
            if g.likes:
                t *= -1
        
        return t
    
    def __contains__(self, student):
        return student in self.aff_changes
    
    def log(self):
        """
        prints our student's reactions in a readable form, eg:
        "Helen decides she likes speaker Mike LESS now (-1)"
        """
        for s in self.aff_changes:
            change = self.aff_changes[s]
            if change == 0:
                continue
            elif change < 0:
                mod = 'LESS'
            elif change > 0:
                mod = 'MORE'
            # + to indicate positive change
            if change > 0:
                change = '+' + str(change)
            if log: print "%s REACTS: %s likes %s %s now (%s)" % (self.student,
                self.student.subject_pronoun(), s.name, mod, change)

    def thoughts(self):
        """
        returns a string from 'student thoughts' dict that
        reflect reaction to our gossip
        """
        sub_aff = self.student.affinity[self.gossip.subject]
        obj_aff = self.student.affinity[self.gossip.object]
        
        # FIXME: this logic is grody, figure out a simpler way!
        # we like subject and object equally - change both
        if abs(sub_aff) == abs(obj_aff):
            if sub_aff >= 0:
                sub = 'friend'
            else:
                sub = 'enemy'
            if obj_aff >= 0:
                obj = 'friend'
            else:
                obj = 'enemy'
        # subject is changing our opinion of object
        elif abs(sub_aff) > abs(obj_aff):
            if sub_aff >= 0:
                sub = 'best friend'
            else:
                sub = 'worst enemy'
            if obj_aff >= 0:
                obj = 'friend'
            else:
                obj = 'enemy'
        # object is changing our opinion of subject
        else:
            if sub_aff >= 0:
                sub = 'friend'
            else:
                sub = 'enemy'
            if obj_aff >= 0:
                obj = 'best friend'
            else:
                obj = 'worst enemy'
        if self.gossip.subject is self.student:
            sub = 'me'
        elif self.gossip.object is self.student:
            obj = 'me'
        if self.gossip.likes:
            verb = 'likes'
        else:
            verb = 'hates'
        rstr = '%s %s %s' % (sub, verb, obj)
        if rstr in thoughts.gut_reactions:
            thought_list = thoughts.gut_reactions[rstr]
            reac = sample(thought_list, 1)[0]
        else:
            print "reaction_thoughts: didn't have a string for: %s" % rstr
            return ''
        reac = thoughts.parse(reac, self.gossip)
        if len(reac) > 0 and not reac[0].isupper():
            reac[0] = reac[0].upper()
        return reac


def change_mutual_affinity(group_a, group_b, change=CLIQUE_AFFINITY_BUMP, variance=CLIQUE_AFFINITY_VARIANCE):
    """
    changes the affinity of group A for B and of B for A by the specified
    amount, with the specified range of variance
    """
    for student in group_a:
        for other in group_b:
            if other is student:
                continue
            aff = change + randint(-variance, variance)
            student.affinity[other] += aff
            other.affinity[student] += aff
            if log: print "changing %s's affinity for %s by %s to %s" % (student, other, aff, student.affinity[other])


def list_students(student_list, heading='students', aff_student=None):
    "returns a properly formatted list of students with specified heading"
    s = '%s: ' % heading
    for student in student_list:
        # if aff_student is supplied, output their affinity for students in list
        if aff_student:
            s += "%s (%s), " % (student.name, aff_student.affinity[student])
        else:
            s += "%s, " % student.name
    return s[:-2]


# unintentional political commentary
MALE = 0
FEMALE = 1

class Student():
    
    self_esteem = DEFAULT_SELF_ESTEEM
    
    def __init__(self, x, y):
        self.face = ui.StudentFace(cui, self, x, y)
        # FIXME: this doesn't distribution student names properly
        # for some values of rows/columns
        self.num = (DESK_ROWS + 1) * y + x
        self.name = student_names[self.num]
        # student name list alternates male / female to make
        # gender assignments simple
        self.gender = FEMALE
        if self.num % 2:
            self.gender = MALE
        self.affinity = {}
        is_player = False
        self.reaction = None
        self.clique = None
        self.generate_personality()
    
    def __str__(self):
        return self.name
    
    def init_affinity(self, student_list):
        for s in student_list:
            # self-image is a special case
            if s is self:
                self.affinity[s] = self.self_esteem
            else:
                self.affinity[s] = randint(-START_AFFINITY_MAX, START_AFFINITY_MAX)
            # human player gets to decide who they like and don't :)
            if self is player:
                self.affinity[s] = 0
        if log:
            self.log_friends()
            self.log_enemies()
    
    def generate_personality(self):
        "determines student's personality variables - see above for definitions"
        self.forgiveness = randint(-FORGIVENESS_RANGE, FORGIVENESS_RANGE)
        self.fickleness = randint(-FICKLENESS_RANGE, FICKLENESS_RANGE)
        self.gullibility = randint(-GULLIBILITY_RANGE, GULLIBILITY_RANGE)
    
    def total_popularity(self):
        p = 0
        for s in students:
            if s is self:
                continue
            p += s.affinity[self]
        return p
    
    def log_friends(self):
        fl = self.friends()
        if len(fl) == 0:
            print "%s has no friends :(" % self.name
            return
        header = "%s's friend(s)" % self.name
        print list_students(fl, header, self)
        bfl = self.best_friends()
        header = "%s's best friend(s)" % self.name
        print list_students(bfl, header, self)
    
    def friends(self):
        "returns a list of our friends, sorted from least to highest affinity"
        fl = []
        for s in self.affinity:
            if s is self:
                continue
            if self.affinity[s] <= 0:
                continue
            # positive affinity = friend
            fl.append(s)
        # python is awesome
        fl.sort(cmp=lambda x,y: cmp(self.affinity[x], self.affinity[y]))
        return fl
    
    def best_friends(self):
        "returns a list of our best-liked friends"
        bfl = []
        best_aff = 0
        for s in self.friends():
            if self.affinity[s] == best_aff:
                bfl.append(s)
            elif self.affinity[s] > best_aff:
                best_aff = self.affinity[s]
                bfl = [s]
        return bfl
    
    def log_enemies(self):
        el = self.enemies()
        if len(el) == 0:
            print "%s has no enemies :D" % self.name
            return
        header = "%s's enemies(s)" % self.name
        print list_students(el, header, self)
        wel = self.worst_enemies()
        header = "%s's worst enemies(s)" % self.name
        print list_students(wel, header, self)
    
    def enemies(self):
        "same as Student.friends() but for enemies"
        el = []
        for s in self.affinity:
            if s is self:
                continue
            if self.affinity[s] >= 0:
                continue
            # negative affinity = enemy
            el.append(s)
        el.sort(cmp=lambda x,y: cmp(self.affinity[x], self.affinity[y]))
        # worst enemies at the end of the list
        el.reverse()
        return el
    
    def worst_enemies(self):
        wel = []
        worst_aff = 0
        for s in self.enemies():
            if self.affinity[s] == worst_aff:
                wel.append(s)
            elif self.affinity[s] < worst_aff:
                worst_aff = self.affinity[s]
                wel = [s]
        return wel
    
    def log_affinity(self, gossip):
        """
        prints this student's affinity for the people involved in
        specified gossip in a readable form, eg:
        "Helen LIKES speaker Mike (3)"
        """
        for s, title in [(gossip.speaker, 'speaker'),
            (gossip.subject, 'subject'), (gossip.object, 'object')]:
            aff = self.affinity[s]
            if aff > 0:
                verb = 'LIKES'
            elif aff < 0:
                verb = 'DISLIKES'
            else:
                verb = 'is MEH about'
            print "%s %s %s %s (%s)" % (self.name, verb, title, s.name, aff)
    
    def subject_pronoun(self):
        "returns this student's subject pronoun, ie he/she"
        if self.gender == MALE:
            return 'he'
        else:
            return 'she'
    
    def object_pronoun(self):
        if self.gender == MALE:
            return 'him'
        else:
            return 'her'
    
    def possess_pronoun(self):
        if self.gender == MALE:
            return 'his'
        else:
            return 'her'
    
    def apply_reaction(self, reaction):
        """
        commits the values in the specified gossip-reaction class
        to this student's affinity map
        (separated from generating the reaction so that we can
        preview gossip results without committing to them)
        """
        for s in reaction.aff_changes:
            self.affinity[s] += reaction.aff_changes[s]
        # any time the affinity map changes, update which clique we're in
        update_cliques()
    
    def update(self):
        # student faces get updated from cui.update()
        
        # only generate a new reaction if we're hearing new gossip
        # (ie don't recalc every frame and spam the log)
        if cui.gossip:
            if not self.reaction or (self.reaction.gossip is not cui.gossip):
                self.reaction = Reaction(self, cui.gossip)


class Clique():
    """
    a group of friends created before game start to pre-seed the social network
    
    cliques help organize the students into semi-factiony groups
    with natural lines of affinity that help the player form social intuitions
    (when affinity is non-mutual and scattered the "terrain" is much harder
    to read)
    """
    
    exclusivity = CLIQUE_EXCLUSIVITY
    
    def __init__(self):
        # even if we fail to form, student list length will be checked
        self.students = []
        # determine group size before we generate clique's identity
        group_size = randint(2, CLIQUE_MAX_SIZE)
        # form new cliques out of students who are not in existing ones
        # (account for "not in cliques" list having shrunk < intended size)
        group_size = min(len(cm.students_not_in_cliques), group_size)
        # don't form a clique if it would be too small
        if group_size < 2:
            return
        # clique identity
        # use even distribution color gen
        self.color = cycled_color_with_value()
        # name: fictional basis for clique
        # get from pool of names in CliqueManager
        self.name = cm.get_new_name()
        #
        # build students list from group
        #
        group = sample(cm.students_not_in_cliques, group_size)
        for student in group:
            self.students.append(student)
            student.clique = self
            cm.students_not_in_cliques.remove(student)
        # create mutual affinity between every student in this clique,
        # with some minor variation
        change_mutual_affinity(self.students, self.students)
        # cliques should often (but not always) be mutually antagonistic
        if cm.cliques:
            # pick a random clique
            enemy_clique = sample(cm.cliques, 1)[0]
            change_mutual_affinity(self.students, enemy_clique.students, -CLIQUE_AFFINITY_BUMP)
        self.sort_by_popularity()
        if log: print self
    
    def __str__(self):
        # rather than just use self.students, list students with their
        # rank within this clique
        sl = 'clique: %s: ' % self.name
        for s in self.students:
            sl += '%s (%s), ' % (s, self.aggregate_affinity(s))
        sl = sl[:-2]
        return sl
    
    def sort_by_popularity(self):
        """
        sort student list by aggregate affinity, so that last in list
        is most popular
        """
        self.students.sort(cmp=lambda x,y: cmp(self.aggregate_affinity(x), self.aggregate_affinity(y)))
    
    def most_popular(self):
        "returns the student most popular within this clique"
        #return self.students[-1]
        most_pop = None
        highest_aff = 0
        for s in self.students:
            aff = self.aggregate_affinity(s)
            if aff > highest_aff:
                highest_aff = aff
                most_pop = s
        return most_pop
    
    def check_disband(self):
        """
        if membership has dropped to 1 or 0, disbands this clique and returns
        True
        """
        if len(self.students) < 2:
            print "%s is disbanding" % self
            for s in self.students:
                s.clique = None
            cliques.remove(self)
            return True
        return False
    
    def rank(self, student):
        "returns this student's normalized (1.0) rank within this clique"
        # (1.0 = most popular)
        if student not in self.students:
            return 0.0
        aff = self.aggregate_affinity(student)
        highest_aff = self.aggregate_affinity(self.most_popular())
        r = float(aff) / highest_aff
        #print "%s's rank within %s: %s" % (student, self, r)
        return r
    
    def average_affinity(self, student):
        "returns a student's average affinity with this clique"
        agg_aff = float(self.aggregate_affinity(student))
        # if student list is too short, we are probably due for disbandment
        if len(self.students) < 2:
            return 0
        # for an accurate average, pretend we're not in this clique
        if student in self.students:
            avg_aff = agg_aff / (len(self.students) - 1)
        else:
            avg_aff = agg_aff / len(self.students)
        return avg_aff
    
    def aggregate_affinity(self, student):
        "returns the specified student's aggregate affinity with this clique"
        aff = 0
        for other in self.students:
            # self-esteem never counts towards aggregate affinity
            if student is not other:
                aff += other.affinity[student]
        return aff


class CliqueManager():
    
    "creates and updates the list of cliques"
    
    # copy[:] this to an instance var so we don't run out on multiple init()s
    clique_names = [
        'band class',
        'varsity basketball',
        'computer club',
        'drama department',
        'football team',
        'chess club',
        'glee club',
        'debate club',
        'ping pong team',
        'swim team',
        'volleyball team',
        'detention'
    ]
    
    def __init__(self):
        "basic startup, rest is in self.build_clique_list"
        self.cliques = []
        self.available_clique_names = self.clique_names[:]
        # player shouldn't be in any clique
        self.students_not_in_cliques = students[:]
        self.students_not_in_cliques.remove(player)
    
    def build_clique_list(self):
        """
        populate self.cliques and finish init
        (can't do this in __init__ because our instance needs
        to exist when cliques are being created)
        """
        for i in range(INITIAL_CLIQUES):
            c = Clique()
            if len(c.students) > 0:
                self.cliques.append(c)
        if self.students_not_in_cliques:
            print list_students(self.students_not_in_cliques, 'students not in cliques')
        # generate initial clique colors
        for s in students:
            s.face.refresh_clique_color()
        self.update_cliques()
    
    def get_new_name(self):
        "returns an unused name from our pool of them"
        n = sample(self.available_clique_names, 1)[0]
        self.available_clique_names.remove(n)
        return n
    
    def update_cliques(self):
        "re-evaluate all clique memberships"
        changed_students = []
        for s in students:
            for c in self.cliques:
                # student has no clique
                if not s.clique:
                    # if we don't have a clique, are we popular enough with
                    # this one to join it?
                    aff = c.average_affinity(s)
                    if aff >= c.exclusivity:
                        # TODO: induction into a clique is a big deal,
                        # put up a UI dialog or something
                        print "%s is now part of %s" % (s, c)
                        c.students.append(s)
                        changed_students.append(s)
                    continue
                # this is student's current clique
                if s.clique is c:
                    # have we slipped below our clique's exclusivity threshold?
                    # if so remove us
                    aff = c.average_affinity(s)
                    if aff < c.exclusivity:
                        # TODO: falling out with a clique is also big deal, UI etc
                        print "%s is no longer part of %s" % (s, c)
                        c.students.remove(s)
                        changed_students.append(s)
                    continue
                # this is a clique other than student's current one
                # don't use aggregate affinity; larger cliques will absorb smaller
                my_aff = s.clique.average_affinity(s)
                new_aff = c.average_affinity(s)
                # if we're now liked more by another, remove from old & add to new
                if my_aff < new_aff and new_aff >= c.exclusivity:
                    print "%s is no longer part of %s" % (s, c)
                    s.clique.students.remove(s)
                    print "%s is now part of %s" % (s, c)
                    c.students.append(s)
                    changed_students.append(s)
        # re-sort
        for c in self.cliques:
            c.sort_by_popularity()
        # check for cliques to disband
        disbands = False
        for c in self.cliques:
            disbands = c.check_disband()
        # re-update if we did disband some cliques
        if disbands:
            self.update_cliques()
        # finally, refresh the clique colors of all students who changed cliques
        for s in changed_students:
            s.face.refresh_clique_color()


def test_reaction_dialog():
    global testing_reaction_dialog
    "puts a random reaction dialog on the screen for easy testing"
    g = random_gossip()
    l = sample(students, 1)[0]
    # generate listener reaction before creating dialog
    l.reaction = Reaction(l, g)
    cui.dialog = ui.ReactionDialog(l.reaction)
    # let ui know we're in a debug situation
    testing_reaction_dialog = True


def input():
    global draw_total_pop, draw_gossip_result_preview, draw_affinity
    global testing_reaction_dialog
    if cui.can_input:
        ui.mouse_x, ui.mouse_y = pygame.mouse.get_pos()
    elif cui.note:
        ui.mouse_x, ui.mouse_y = cui.note.loc.x, cui.note.loc.y
    for event in pygame.event.get():
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                exit()
            if event.key == pygame.K_F1:
                init()
            if event.key == pygame.K_F2:
                ui.draw_total_pop = not ui.draw_total_pop
            if event.key == pygame.K_F3:
                ui.draw_gossip_result_preview = not ui.draw_gossip_result_preview
            if event.key == pygame.K_F4:
                ui.draw_affinity = not ui.draw_affinity
            if event.key == pygame.K_F5:
                test_reaction_dialog()
        elif event.type == pygame.MOUSEBUTTONDOWN:
            if cui.can_input:
                cui.click()
            # let us dismiss test dialogs easily
            if testing_reaction_dialog:
                testing_reaction_dialog = False
                cui.dialog = None
        elif event.type == pygame.MOUSEBUTTONUP:
            if cui.can_input:
                cui.unclick()
        # support close button in windows
        elif event.type == pygame.QUIT:
            exit()


def update():
    for s in students:
        s.update()
    cui.update()


def draw():
    screen.fill(NEAR_BLACK)
    cui.draw()
    pygame.display.flip()


def main_loop():
    global time
    input()
    update()
    draw()
    time += clock.tick_busy_loop(FRAMERATE)
    ui.time = time


def init():
    global time, cui
    global students, player, students_not_in_cliques, cm
    if log: print "Initializing..."
    time = 0
    ui.init()
    # fill in some vars for ui submodule
    # (is this stupid?)
    ui.log = log
    ui.screen = screen
    ui.debug_font = debug_font
    ui.debug_font_small = debug_font_small
    cui = ui.ClassroomUI()
    # (this is DEFINITELY stupid)
    ui.ui = cui
    # init sim objects
    students = []
    player = None
    if log: print "populating classroom..."
    # create list of students
    for row in range(DESK_ROWS):
        for col in range(DESK_COLUMNS):
            students.append(Student(col, row))
    # player sits in "middle"
    player = students[PLAYER_DESK - 1] # account for base-0-ness
    ui.students = students
    ui.player = player
    if log: print "player is %s" % player.name
    # populate students' affinity maps
    for s in students:
        s.init_affinity(students)
    if log: print "forming cliques..."
    # organize students into "friend groups" ie cliques
    cm = CliqueManager()
    cm.build_clique_list()
    # populate UI with visible elements of students
    for s in students:
        cui.elements.append(s.face)
    # simulate a # of turns to vary affinity map at game start
    for i in range(START_AFFINITY_ITERATIONS):
        g = random_gossip()
        print g
        # each student processes this gossip
        for s in students:
            r = s.reaction(g)
            s.log_reaction(r)
            s.apply_reaction(r)
    # make window icon player's face
    player_icon = player.face.face_style.scaled_thumbnail(0.25, 0.25)
    pygame.display.set_icon(player_icon)


if __name__ == "__main__":
    pygame.init()
    clock = pygame.time.Clock()
    if FULLSCREEN:
        screen = pygame.display.set_mode(SCREEN_SIZE, pygame.FULLSCREEN)
    else:
        screen = pygame.display.set_mode(SCREEN_SIZE)
    pygame.display.set_caption(ui.TITLE_TEXT)
    # game data
    df = pygame.font.get_default_font()
    debug_font = pygame.font.Font(df, FONT_SIZE)
    debug_font_small = pygame.font.Font(df, SMALL_FONT_SIZE)
    init()
    while True:
        main_loop()
