# reaction prototype 1

"""
listener face on middle of far left

students who listener likes most are very near (mercury) while hated enemies are far away (pluto)

when reaction is shown, students change their orbital distance

have each reaction member appear, listener shows affinity-face for each, then show changes

---

listener
tracks mouse (or other?)
has affinity for all other students
changes expression (affinity) during reaction

other
draw thumbnail at proper orbit distance
look straight ahead
moves during reaction

---

listener face, neutral, looking right

speaker face appears
listener face: speaker affinity

repeat for subject and object

"""

import pygame
import face
import vector
import lerper
from colors import *
from sys import exit
from random import randint


FULLSCREEN = False
SCREEN_SIZE = SCREEN_WIDTH, SCREEN_HEIGHT = 1024, 768
FRAMERATE = 30

# override scale to use full-res art
# (this is as big as a face will get without upscaling)
face.art_scale = 1.0

# maximum extent of affinity
# (used for screen scale calculations)
AFF_MAX = 10

class ListenerFace():
    
    def __init__(self):
        self.style = feature_dir.random_style()
        self.loc = vector.Vector2D(0, SCREEN_HEIGHT / 2)
        self.loc.y -= head_size.y / 2
        self.affinity = {}
        self.aff = 0
        self.affinity = {}
    
    def draw(self):
        if updating_student:
            self.aff = self.affinity[updating_student]
        else:
            self.aff = 0
        # if a student is updating look at them, else look at mouse
        if updating_student:
            dir = updating_student.loc - self.loc
        else:
            dir = vector.Vector2D(mouse_x, mouse_y) - self.loc
        dir = dir.normal()
        self.style.draw(screen, self.loc, dir, self.aff)


class GossipFace():
    
    scale_rate = 0.05
    move_rate = 20
    
    # TODO: min and max scales (within affinity range)
    
    def __init__(self, name, x=SCREEN_WIDTH/2, y=SCREEN_HEIGHT/2):
        self.name = name
        self.style = feature_dir.random_style()
        start = vector.Vector2D(x, y)
        self.mover = lerper.Lerper2D(start, move_manager)
        self.loc = self.mover.value
        self.scaler = lerper.Lerper(0, scale_manager)
        self.scale_to(0.5)
    
    def refresh_img(self, aff):
        # updates our image with the specified expression (affinity)
        self.img = pygame.Surface(head_size.int_tuple())
        self.style.draw(self.img, vector.ORIGIN, vector.ORIGIN, aff)
    
    def move_to(self, new_loc):
        self.mover.lerp_at_rate(new_loc, self.move_rate)
    
    def scale_to(self, new_scale):
        self.scaler.lerp_at_rate(new_scale, self.scale_rate)
    
    def change_affinity(self, aff):
        lf.affinity[self] = aff
        self.refresh_img(aff)
        self.move_to(vector.Vector2D(screen_x_from_affinity(aff), self.loc.y))
        # base our scale off listener's affinity for us
        scale = abs(aff) + 1 # 0.1 = minimum scale
        scale = min(scale, 9) # maximum scale
        scale /= 10.0
        self.scaler.steps = self.mover.steps
        self.scale_to(scale)
    
    def draw(self):
        loc = self.mover.value
        size = head_size * self.scaler.value
        # loc = center
        x = loc.x - size.x / 2
        y = loc.y - size.y / 2
        face_img = pygame.transform.scale(self.img, size.int_tuple())
        screen.blit(face_img, (x, y))
        # draw name above head
        if updating_student is self:
            return
        name_surf = debug_font.render(self.name, True, LIGHT_GREY)
        y_offset = head_size.y * self.scaler.value * 0.75 - 10
        x_offset = name_surf.get_width() / 2
        screen.blit(name_surf, (loc.x - x_offset, loc.y - y_offset))


def input():
    global mouse_x, mouse_y
    mouse_x, mouse_y = pygame.mouse.get_pos()
    for event in pygame.event.get():
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                exit()
            elif event.key == pygame.K_F1:
                init()
            # test movement / affinity change
            elif event.key == pygame.K_F2:
                subj.change_affinity(randint(-AFF_MAX, AFF_MAX))
        elif event.type == pygame.QUIT:
            exit()


def update():
    global updating_student
    updating_student = None
    scale_manager.update()
    move_manager.update()
    # update scales / locations of each GossipFace in order,
    # skipping to next only if current is at its target scale/loc
    for e in [spk, subj, obj]:
        if e.scaler.updated or e.mover.updated:
            updating_student = e
            break


def draw():
    screen.fill(BLACK)
    for e in [lf, spk, subj, obj]:
        e.draw()
    pygame.display.flip()


def main_loop():
    input()
    update()
    draw()
    clock.tick_busy_loop(FRAMERATE)


LEFT_MARGIN = 200


def screen_x_from_affinity(aff):
    # clamp, leave margin of 1
    if aff <= -AFF_MAX:
        aff = -AFF_MAX + 1
    elif aff >= AFF_MAX:
        aff = AFF_MAX - 1
    # base at 0
    aff -= AFF_MAX
    # flip so high positive affinity is closest, low negative is furthest
    aff = -aff
    # find proportional x
    x = aff * (SCREEN_WIDTH - LEFT_MARGIN)
    x /= AFF_MAX * 2.0
    x += LEFT_MARGIN
    return int(x)


def init():
    global scale_manager, move_manager
    global lf, spk, subj, obj
    scale_manager = lerper.LerperManager()
    scale_manager.one_at_a_time = True
    move_manager = lerper.LerperManager()
    move_manager.one_at_a_time = True
    #move_manager.log = True
    lf = ListenerFace()
    # gossip speaker
    y = SCREEN_HEIGHT / 4
    gossipers = []
    for s in ['speaker', 'subject', 'object']:
        aff = randint(-AFF_MAX, AFF_MAX)
        student = GossipFace(s, screen_x_from_affinity(aff), y)
        lf.affinity[student] = aff
        # student's expression = listener affinity for student
        # (even if it's not the student's actual affinity for listener)
        student.refresh_img(aff)
        y += SCREEN_HEIGHT / 4
        gossipers.append(student)
    # set globals
    spk = gossipers[0]
    subj = gossipers[1]
    obj = gossipers[2]
    # TODO: generate reaction: affinity changes for spk, subj, obj
    # move and scale gossipfaces accordingly, show listener expression


if __name__ == "__main__":
    pygame.init()
    clock = pygame.time.Clock()
    if FULLSCREEN:
        screen = pygame.display.set_mode(SCREEN_SIZE, pygame.FULLSCREEN)
    else:
        screen = pygame.display.set_mode(SCREEN_SIZE)
    pygame.display.set_caption('POPULAR reaction prototype 1')
    pygame.key.set_repeat(50)
    df = pygame.font.get_default_font()
    debug_font = pygame.font.Font(df, 16)
    feature_dir = face.FaceFeatureDirectory()
    head_size = vector.Vector2D(feature_dir.heads[0].stages[0].image.get_size())
    init()
    while True:
        main_loop()
