# "student thoughts" submodule
from rangedict import RangeDictionary

# access thoughts via a 'reference string',
# derived from details of some gossip

# markup guide:
# [SPK_NAME] - speaker's name
# [SPK_OBJ] - speaker's object pronoun, ie him/her
# [SUB] - subject's subject pronoun, ie he/she
# [OBJ] - object's subject pronoun, same
# [SUB_NAME] - subject's name
# [OBJ_NAME] - object's name
# [SUB_OBJ] - subject's object pronoun, ie him/her
# [OBJ_OBJ] - object's object pronoun, same
# [SUB_POS] - subject's possessive pronoun, ie his/her
# [OBJ_POS] - object's possessive pronoun, same

gut_reactions = {

# 9 likes 1
'best friend likes friend':
[
"Cool - pals think alike!"
],

# 9 likes -1
'best friend likes enemy':
[
"Well, I guess [OBJ_NAME] isn't so bad."
],

# 1 likes 9
'friend likes best friend':
[
"Cool, [SUB_NAME] has good taste."
],

# 1 likes -9
'friend likes worst enemy':
[
"I can't believe it - how could [SUB_NAME] be so lame?"
],

# -1 likes 9
'enemy likes best friend':
[
"Well, [SUB_NAME] has good taste, [SUB] can't be all that bad."
],

# -1 likes -9
'enemy likes worst enemy':
[
"Hmph - rotten likes rotten."
],

# -9 likes 1
'worst enemy likes friend':
[
"Wha?  I thought [OBJ_NAME] was okay but clearly not!"
],

# -9 likes -1
'worst enemy likes enemy':
[
"I knew it... [OBJ]'s pals with evil incarnate."
],

# 9 hates 1
'best friend hates friend':
[
"[OBJ_NAME] must be bad... guess I was wrong about [OBJ_OBJ]!"
],

# 9 hates -1
'best friend hates enemy':
[
"OMG - [OBJ_NAME] is even worse than I thought!"
],

# 1 hates 9
'friend hates best friend':
[
"Screw [SUB_NAME] - nobody acts that way to my best pal!"
],

# 1 hates -9
'friend hates worst enemy':
[
"Right on, [SUB_NAME] - [OBJ_NAME] is scum!"
],

# -1 hates 9
'enemy hates best friend':
[
"Sheesh - [OBJ_NAME] is even worse than I thought!"
],

# -1 hates -9
'enemy hates worst enemy':
[
"Well, I can't blame [SUB_OBJ] for hating [OBJ_NAME]."
],

# -9 hates 1
'worst enemy hates friend':
[
"That rotten jerk [SUB_NAME] hates ALL my friends."
],

# -9 hates -1
'worst enemy hates enemy':
[
"Maybe [OBJ_NAME] isn't so bad, if evil incarnate hates [OBJ_OBJ]."
],

# 5 likes 5
'friend likes friend':
[
"Cool, all my friends get along."
],

# 5 likes -5
'friend likes enemy':
[
"Hmm, maybe I was wrong about [OBJ_NAME]."
],

# -5 likes -5
'enemy likes enemy':
[
"Screw 'em both, then."
],

# -5 likes 5
'enemy likes friend':
[
"Hmm, maybe I was wrong about [SUB_NAME]."
],

# 5 hates 5
'friend hates friend':
[
"What's [SUB_NAME]'s deal?  Why [SUB] gotta cause trouble?"
],

# 5 hates -5
'friend hates enemy':
[
"Right on - [OBJ_NAME] is a jerk."
],

# -5 hates -5
'enemy hates enemy':
[
"Well, I guess even jerks like [SUB_NAME] have good sense sometimes."
],

# -5 hates 5
'enemy hates friend':
[
"Pff, typical."
],

# I like a friend
'me likes friend':
[
"Well sure, of course I like [OBJ_NAME]."
],

# I like an enemy
'me likes enemy':
[
"What's all this?  I don't like [OBJ_NAME]!"
],

# I hate a friend
'me hates friend':
[
"No way!  [OBJ_NAME] is my buddy!"
],

# I hate an enemey
'me hates enemy':
[
"DUH.  Of course I hate [OBJ_NAME]."
],

# a friend likes me
'friend likes me':
[
"You're pretty cool too, [SUB_NAME]!"
],

# an enemy likes me
'enemy likes me':
[
"Wha?  I wouldn't have expected that from [SUB_NAME]."
],

# a friend hates me
'friend hates me':
[
"GAH!  How could [SUB_NAME] possibly think that about me?!?"
],

# an enemy hates me
'enemy hates me':
[
"The feeling is mutual... [SUB_NAME] is a loser."
]

}

# some permutations are redundant, fill them in with existing values
gut_reactions['best friend likes me'] = gut_reactions['friend likes me']
gut_reactions['best friend hates me'] = gut_reactions['friend hates me']
gut_reactions['worst enemy likes me'] = gut_reactions['enemy likes me']
gut_reactions['worst enemy hates me'] = gut_reactions['enemy hates me']
gut_reactions['me likes best friend'] = gut_reactions['me likes friend']
gut_reactions['me hates best friend'] = gut_reactions['me hates friend']
gut_reactions['me likes worst enemy'] = gut_reactions['me likes enemy']
gut_reactions['me hates worst enemy'] = gut_reactions['me hates enemy']


#
# very short summaries of reaction affinity changes
#
# TODO: these kinda suck, are they even useful and/or can they be redone?
reaction_summary = {
    5:'OMG!',
    4:'Awesome!',
    3:'Great!',
    2:'Cool!',
    1:'Yay!',
    0:'Meh.',
    -1:'Hmph.',
    -2:'Pfft.',
    -3:'Grr.',
    -4:'JERK!',
    -5:'@#*%!!'
}

# thoughts about speakers of various affinities
speaker_affinities = RangeDictionary({
    # TODO: zero affinity might mean "don't know, don't care"
    # OR it might mean "crossing over between like/dislike" -
    # detect and handle these cases specifically?
    (0, 0): 'Who is [SPK_NAME] again?',
    (0, 4): '[SPK_NAME] is okay, I guess I believe [SPK_OBJ].',
    (4, 8): '[SPK_NAME] is cool, I believe [SPK_OBJ].',
    (8, 1000): '[SPK_NAME] is awesome, I totally believe [SPK_OBJ]!',
    (-4, 0): '[SPK_NAME] is kinda lame, I dunno if I believe [SPK_OBJ].',
    (-8, -4): '[SPK_NAME] is a loser, why should I believe [SPK_OBJ]?',
    (-1000, -8): '[SPK_NAME] is a total jerk, no way I believe [SPK_OBJ]!'
})

# thoughts about the truthiness of a statement
# TODO: how to determine this, basic form of the statements,
# vars it needs access to

# in general:
# truthy things are "nice"
# untruthy things are "mean"

truthiness = {
    0:0
    # anyone hates friend
    # anyone hates listener
    # """
    # "I don't believe that... you're mean!"
    # "That's not true - what a jerky thing to say!"
    # # anyone likes friend
    # # anyone likes listener (ie listener is gossip.object)
    # "Sure, yeah... that's a nice thing to say."
    # # enemy/neutral hates another enemy
    # # enemy likes a friend
    # "I dunno about that, but it's a nice thought."
    # """
}


affinity_descriptions = RangeDictionary({
    (0, 0): 'Acquaintance',
    (0, 3): 'Friend',
    (3, 6): 'Good Friend',
    (6, 1000): 'Close Friend',
    (-3, 0): 'Enemy',
    (-6, -3): 'Hated Enemy',
    (-1000, -6): 'Despised Enemy'
})


def describe_affinity(student, other):
    return affinity_descriptions[student.affinity[other]]


def parse(thought, gossip):
    "parses the given string for appropriate articles based on gossip"
    spk = gossip.speaker
    sub = gossip.subject
    obj = gossip.object
    thought = thought.replace('[SPK_NAME]', spk.name)
    thought = thought.replace('[SPK_OBJ]', spk.object_pronoun())
    thought = thought.replace('[SUB]', sub.subject_pronoun())
    thought = thought.replace('[OBJ]', obj.subject_pronoun())
    thought = thought.replace('[SUB_NAME]', sub.name)
    thought = thought.replace('[OBJ_NAME]', obj.name)
    thought = thought.replace('[SUB_OBJ]', sub.object_pronoun())
    thought = thought.replace('[OBJ_OBJ]', obj.object_pronoun())
    thought = thought.replace('[SUB_POS]', sub.possess_pronoun())
    thought = thought.replace('[OBJ_POS]', obj.possess_pronoun())
    return thought
